from django.contrib import admin

from .models import JWTSecret


@admin.register(JWTSecret)
class JWTSecretAdmin(admin.ModelAdmin):
    list_display = ('identifier',)
    search_fields = ('identifier',)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse


class APIMixin:

    def get_absolute_api_url(self, request=None, **kwargs) -> str:
        """
        Build the absolute URL of the object in the API.
        """
        # build the URL of the informatieobject
        resource_name = self._meta.model_name

        reverse_kwargs = {'uuid': self.uuid}
        reverse_kwargs.update(**kwargs)

        url = reverse(
            f'{resource_name}-detail',
            kwargs=reverse_kwargs, request=request
        )
        return url


class JWTSecret(models.Model):
    identifier = models.CharField(_("identifier"), max_length=50, unique=True)
    secret = models.CharField(_("secret"), max_length=255)

    def __str__(self):
        return self.identifier



import logging
from collections import OrderedDict

from django.conf import settings

from drf_yasg import openapi
from drf_yasg.inspectors import SwaggerAutoSchema
from rest_framework import exceptions, serializers, status

from ..exceptions import Conflict, Gone, PreconditionFailed
from ..geo import GeoMixin
from ..permissions import ActionScopesRequired
from ..search import is_search_view
from ..serializers import FoutSerializer, ValidatieFoutSerializer

logger = logging.getLogger(__name__)

TYPE_TO_FIELDMAPPING = {
    openapi.TYPE_INTEGER: serializers.IntegerField,
    openapi.TYPE_NUMBER: serializers.FloatField,
    openapi.TYPE_STRING: serializers.CharField,
    openapi.TYPE_BOOLEAN: serializers.BooleanField,
}

COMMON_ERRORS = [
    exceptions.AuthenticationFailed,
    exceptions.NotAuthenticated,
    exceptions.PermissionDenied,
    exceptions.NotAcceptable,
    Conflict,
    Gone,
    exceptions.UnsupportedMediaType,
    exceptions.Throttled,
    exceptions.APIException,
]


DEFAULT_ACTION_ERRORS = {
    'create': COMMON_ERRORS + [
        exceptions.ParseError,
        exceptions.ValidationError,
    ],
    'list': COMMON_ERRORS,
    'retrieve': COMMON_ERRORS + [
        exceptions.NotFound,
    ],
    'update': COMMON_ERRORS + [
        exceptions.ParseError,
        exceptions.ValidationError,
        exceptions.NotFound,
    ],
    'partial_update': COMMON_ERRORS + [
        exceptions.ParseError,
        exceptions.ValidationError,
        exceptions.NotFound,
    ],
    'destroy': COMMON_ERRORS + [
        exceptions.NotFound,
    ],
}


class AutoSchema(SwaggerAutoSchema):

    @property
    def model(self):
        qs = self.view.get_queryset()
        return qs.model

    @property
    def _is_search_view(self):
        return is_search_view(self.view)

    def get_operation_id(self, operation_keys):
        """
        Simply return the model name as lowercase string, postfixed with the operation name.
        """
        action = operation_keys[-1]
        model_name = self.model._meta.model_name
        return f"{model_name}_{action}"

    def should_page(self):
        if self._is_search_view:
            return hasattr(self.view, 'paginator')
        return super().should_page()

    def get_request_serializer(self):
        if not self._is_search_view:
            return super().get_request_serializer()

        Base = self.view.get_search_input_serializer_class()

        filter_fields = []
        for filter_backend in self.view.filter_backends:
            filter_fields += self.probe_inspectors(
                self.filter_inspectors,
                'get_filter_parameters', filter_backend()
            ) or []

        filters = {}
        for filter_field in filter_fields:
            FieldClass = TYPE_TO_FIELDMAPPING[filter_field.type]
            filters[filter_field.name] = FieldClass(
                help_text=filter_field.description,
                required=filter_field.required
            )

        SearchSerializer = type(Base.__name__, (Base,), filters)
        return SearchSerializer()

    def _get_search_responses(self):
        response_status = status.HTTP_200_OK
        response_schema = self.serializer_to_schema(self.get_view_serializer())
        schema = openapi.Schema(type=openapi.TYPE_ARRAY, items=response_schema)
        if self.should_page():
            schema = self.get_paginated_response(schema) or schema
        return OrderedDict({str(response_status): schema})

    def _get_error_responses(self) -> OrderedDict:
        """
        Add the appropriate possible error responses to the schema.

        E.g. - we know that HTTP 400 on a POST/PATCH/PUT leads to validation
        errors, 403 to Permission Denied etc.
        """
        responses = {}

        action = self.view.action

        if action not in DEFAULT_ACTION_ERRORS and self._is_search_view:  # similar to a CREATE
            action = 'create'

        exception_klasses = DEFAULT_ACTION_ERRORS.get(action)
        if exception_klasses is None:
            logger.debug("Unknown action %s, no default error responses added")
            return responses

        fout_schema = self.serializer_to_schema(FoutSerializer())
        for exception_klass in exception_klasses:
            status_code = exception_klass.status_code
            responses[status_code] = fout_schema

        has_validation_errors = any(
            issubclass(klass, exceptions.ValidationError)
            for klass in exception_klasses
        )
        if has_validation_errors:
            schema = self.serializer_to_schema(ValidatieFoutSerializer())
            responses[exceptions.ValidationError.status_code] = schema

        if isinstance(self.view, GeoMixin):
            status_code = PreconditionFailed.status_code
            responses[status_code] = fout_schema

        # sort by status code
        return OrderedDict([
            (status_code, schema)
            for status_code, schema
            in sorted(responses.items())
        ])

    def get_default_responses(self) -> OrderedDict:
        if self._is_search_view:
            responses = self._get_search_responses()
            serializer = self.get_view_serializer()
        else:
            responses = super().get_default_responses()
            serializer = self.get_request_serializer() or self.get_view_serializer()

        # inject any headers
        _responses = OrderedDict()
        for status_, schema in responses.items():
            custom_headers = self.probe_inspectors(
                self.field_inspectors, 'get_response_headers',
                serializer, {'field_inspectors': self.field_inspectors},
                status=status_
            ) or None

            assert isinstance(schema, openapi.Schema.OR_REF) or schema == ''
            response = openapi.Response(
                description='',
                schema=schema or None,
                headers=custom_headers
            )
            _responses[status_] = response

        for status_code, response in self._get_error_responses().items():
            _responses[status_code] = response

        return _responses

    def add_manual_parameters(self, parameters):
        base = super().add_manual_parameters(parameters)
        if self._is_search_view:
            serializer = self.get_request_serializer()
        else:
            serializer = self.get_request_serializer() or self.get_view_serializer()
        extra = self.probe_inspectors(
            self.field_inspectors, 'get_request_header_parameters',
            serializer, {'field_inspectors': self.field_inspectors}
        ) or []
        return base + extra

    def get_security(self):
        """Return a list of security requirements for this operation.

        Returning an empty list marks the endpoint as unauthenticated (i.e. removes all accepted
        authentication schemes). Returning ``None`` will inherit the top-level secuirty requirements.

        :return: security requirements
        :rtype: list[dict[str,list[str]]]"""
        permissions = self.view.get_permissions()
        scope_permissions = [perm for perm in permissions if isinstance(perm, ActionScopesRequired)]

        if not scope_permissions:
            return super().get_security()

        if len(permissions) != len(scope_permissions):
            logger.warning(
                "Can't represent all permissions in OAS for path %s and method %s",
                self.path, self.method
            )

        required_scopes = []
        for perm in scope_permissions:
            scopes = perm.get_required_scopes(self.view)
            if scopes is None:
                continue
            required_scopes.append(scopes)

        if not required_scopes:
            return None  # use global security

        scopes = [str(scope) for scope in sorted(required_scopes)]

        # operation level security
        return [{
            settings.SECURITY_DEFINITION_NAME: scopes,
        }]



import os
import time
from urllib.parse import urlparse

from django.conf import settings

import jwt
import yaml

from .models import JWTSecret

DEFAULT_PATH_PARAMETERS = {
    'version': '1',
}

SPEC_PATH = os.path.join(settings.BASE_DIR, 'src', 'openapi.yaml')

with open(SPEC_PATH, 'r') as infile:
    SPEC = yaml.load(infile)


def get_operation_url(operation, **kwargs):
    url = SPEC['servers'][0]['url']
    base_path = urlparse(url).path

    for path, methods in SPEC['paths'].items():
        for name, method in methods.items():
            if name == 'parameters':
                continue

            if method['operationId'] == operation:
                format_kwargs = DEFAULT_PATH_PARAMETERS.copy()
                format_kwargs.update(**kwargs)
                path = path.format(**format_kwargs)
                return f"{base_path}{path}"

    raise ValueError(f"Operation {operation} not found")


class TypeCheckMixin:

    def assertResponseTypes(self, response_data: dict, types: tuple):
        """
        Do type checks on the response data.

        :param types: tuple of (field_name, class)
        :raises AssertionError: if the type mismatches
        """
        for field, type_ in types:
            with self.subTest(field=field, type=type_):
                self.assertIsInstance(response_data[field], type_)


def get_validation_errors(response, field, index=0):
    """
    Assumes there's only one validation error for the field.
    """
    assert response.status_code == 400
    i = 0
    for error in response.data['invalid-params']:
        if error['name'] != field:
            continue

        if i == index:
            return error

        i += 1


def generate_jwt(scopes: list, secret: str='letmein', zaaktypes: list=None) -> str:
    scope_labels = sum((_get_scope_labels(scope) for scope in scopes), [])
    payload = {
        # standard claims
        'iss': 'testsuite',
        'iat': int(time.time()),
        # custom claims
        'zds': {
            'scopes': scope_labels,
            'zaaktypes': zaaktypes or [],
        },
    }
    headers = {
        'client_identifier': 'testsuite',
    }
    encoded = jwt.encode(payload, secret, headers=headers, algorithm='HS256')
    return encoded


def _get_scope_labels(scope) -> list:
    if not scope.children:
        return [scope.label]

    labels = []
    for child in scope.children:
        labels += _get_scope_labels(child)
    return sorted(set(labels))


class JWTScopesMixin:

    scopes = None
    zaaktypes = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier='testsuite',
            defaults={'secret': 'letmein'}
        )

    def setUp(self):
        super().setUp()

        if self.scopes is not None:
            token = generate_jwt(
                scopes=self.scopes,
                zaaktypes=self.zaaktypes or [],
                secret='letmein'
            )
            self.client.credentials(HTTP_AUTHORIZATION=token)



import logging

from drf_yasg import openapi
from drf_yasg.generators import (
    OpenAPISchemaGenerator as _OpenAPISchemaGenerator
)

logger = logging.getLogger(__name__)


class OpenAPISchemaGenerator(_OpenAPISchemaGenerator):

    def get_path_parameters(self, path, view_cls):
        """Return a list of Parameter instances corresponding to any templated path variables.

        :param str path: templated request path
        :param type view_cls: the view class associated with the path
        :return: path parameters
        :rtype: list[openapi.Parameter]
        """
        parameters = super().get_path_parameters(path, view_cls)

        # see if we can specify UUID a bit more
        for parameter in parameters:
            # the most pragmatic of checks
            if not parameter.name.endswith('_uuid'):
                continue
            parameter.format = openapi.FORMAT_UUID
            parameter.description = "Unieke resource identifier (UUID4)"
        return parameters



import re
from urllib.parse import urlparse

from zds_client.client import UUID_PATTERN


class Response:
    def __init__(self, status_code: int=200):
        self.status_code = status_code


def link_fetcher_404(url: str, *args, **kwargs):
    return Response(status_code=404)


def link_fetcher_200(url: str, *args, **kwargs):
    return Response(status_code=200)


class MockClient:
    data = {}

    @classmethod
    def from_url(cls, detail_url: str, base_dir: str):
        clients = {
            'ztc': ZTCMockClient,
            'drc': DRCMockClient,
        }

        parsed_url = urlparse(detail_url)

        if ':' in parsed_url.netloc:
            host, port = parsed_url.netloc.split(':')
        else:
            host = parsed_url.netloc

        # Try mock client based on host.
        if host in clients:
            return clients[host]()

        # Try mock client based first element of the path.
        first_path_element = parsed_url.path.strip('/').split('/', 1)[0]
        if first_path_element in clients:
            return clients[first_path_element]()

        raise ValueError('Cannot determine service based on url: %s', detail_url)

    def request(self, path: str, operation: str, method='GET', **kwargs):
        bits = path.rsplit('/', 2)

        # Match UUIDs and simple numbers that are most likely used for testing.
        if re.match(r'^({}|[0-9]+)$'.format(UUID_PATTERN), bits[-1]):
            resource = bits[-2]
            uuid = bits[-1]
        else:
            resource = bits[-1]
            uuid = None

        # Should the mock client really consult an OAS to figure out the
        # operation? Workaround for now is to strip off plural forms ("n", "s",
        # "sen").
        if resource not in self.data:
            if resource[0:-1] in self.data:
                resource = resource[0:-1]
            elif resource[0:-3] in self.data:
                resource = resource[0:-3]

        if method == 'GET' and not uuid:
            return self.list(resource)

        return self.retrieve(resource, uuid=uuid)

    def list(self, resource: str, *args, **kwargs):
        return self.data.get(resource, [])

    def retrieve(self, resource: str, *args, **kwargs):
        try:
            index = int(kwargs.get('uuid', 1)) - 1
        except ValueError:
            index = 0

        # result = copy.deepcopy(self.data.get(resource)[index])
        # result['url'] = result['url'].format(**kwargs)
        return self.data.get(resource)[index]


class ZTCMockClient(MockClient):

    data = {
        'statustype': [{
            'url': 'https://ztc/api/v1/catalogussen/{catalogus_uuid}/zaaktypen/{zaaktype_uuid}/statustypen/{uuid}',
            'volgnummer': 1,
            'isEindstatus': False,
        }, {
            'url': 'https://ztc/api/v1/catalogussen/{catalogus_uuid}/zaaktypen/{zaaktype_uuid}/statustypen/{uuid}',
            'volgnummer': 2,
            'isEindstatus': True,
        }],
    }


class DRCMockClient(MockClient):

    data = {
        'objectinformatieobject': [{
            'url': 'https://mock/objectinformatieobjecten/1234',
            'informatieobject': '',
            'object': '',
            'objectType': '',
            'titel': '',
            'beschrijving': '',
            'registratiedatum': '',
        }]
    }


class ObjectInformatieObjectClient(DRCMockClient):
    """
    Kept for backwards compatability.
    """
    @classmethod
    def from_url(cls, *args, **kwargs):
        return cls()

    def list(self, resource, *args, **kwargs):
        assert resource == 'objectinformatieobject'
        return self.data[resource]
